<?php

namespace Drupal\commerce_globalpayments\Plugin\Commerce\PaymentGateway;

use com\realexpayments\remote\sdk\domain\payment\PaymentRequest;
use com\realexpayments\remote\sdk\http\HttpConfiguration;
use com\realexpayments\remote\sdk\RealexClient;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OnsitePaymentGatewayBase;
use Drupal\commerce_payment_example\Plugin\Commerce\PaymentGateway\OnsiteInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Provides base class for Global Payments (Realex) payments.
 */
abstract class GlobalPaymentsBase extends OnsitePaymentGatewayBase implements OnsiteInterface {

  const LIVE_ENDPOINT = 'https://api.realexpayments.com/epage-remote.cgi';
  const TEST_ENDPOINT = 'https://api.sandbox.realexpayments.com/epage-remote.cgi';

  /**
   * Realex SDK API client.
   *
   * @var \com\realexpayments\remote\sdk\RealexClient
   */
  protected $api;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, PaymentTypeManager $payment_type_manager, PaymentMethodTypeManager $payment_method_type_manager, TimeInterface $time) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);

    // Do not initialize payment gateway if it's not configured yet.
    if (empty($configuration['mode']) || empty($configuration['secret'])) {
      return;
    }

    $this->initApi($configuration['mode'], $configuration['secret']);
  }

  /**
   * Prepare Realex SDK.
   *
   * @param string $mode
   *   Payment gateway mode (live or test).
   * @param string $secret
   *   Realex API secret.
   */
  protected function initApi($mode, $secret) {
    if (empty($secret)) {
      return;
    }

    $httpConfig = new HttpConfiguration();
    $endpoint = $mode === 'live' ? self::LIVE_ENDPOINT : self::TEST_ENDPOINT;
    $httpConfig->setEndpoint($endpoint);
    $this->api = new RealexClient($secret);
  }

  /**
   * Makes authenticated Realex API requests using SDK.
   */
  protected function request() {
    return (new PaymentRequest())
      ->addMerchantId($this->configuration['merchant_id'])
      ->addAccount($this->configuration['account_id']);
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'merchant_id' => '',
      'account_id' => '',
      'secret' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['merchant_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Merchant ID'),
      '#default_value' => $this->configuration['merchant_id'],
      '#required' => TRUE,
    ];

    $form['account_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Account ID'),
      '#default_value' => $this->configuration['account_id'],
      '#required' => TRUE,
    ];

    $form['secret'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Shared secret'),
      '#default_value' => $this->configuration['secret'],
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['merchant_id'] = $values['merchant_id'];
      $this->configuration['account_id'] = $values['account_id'];
      $this->configuration['secret'] = $values['secret'];
    }
  }

  /**
   * Format amount for Global Payments API (multiplied by 100).
   *
   * @param float $amount
   *   Amount.
   *
   * @return float|int
   *   Formatted amount.
   */
  public static function formatAmount($amount) {
    return $amount * 100;
  }

  /**
   * Format expiration data for Global Payments API (MMYY).
   *
   * @param string $month
   *   Month number in MM format.
   * @param string $year
   *   Year number in YY format.
   *
   * @return string
   *   Formatted expiry for Realex API.
   */
  public static function formatExpiry($month, $year) {
    return $month . substr($year, -2);
  }

}
